#include "mainwindow.h"
#include "ui_mainwindow.h"


MainWindow::MainWindow(QApplication *a, int screen_size, bool *ran) :
    QMainWindow(),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    m_a = a;
    m_ran = ran;
    m_plotShowing = false;

    QMenu *help;
    QFont menuFont(QString("Noto Sans"), 32);
    help = ui->menuBar->addMenu(QString("&Help"));
    help->setFixedWidth(screen_size / 2);
    help->setFont(menuFont);
    QAction *aboutAction = new QAction(QString("&About"), this);
    help->addAction(aboutAction);
    QAction *closeAction = new QAction(QString("E&xit"), this);
    help->addAction(closeAction);

    QFile fh(QString(":/res/LicenseNotification.txt"));
    fh.open(QFile::ReadOnly | QFile::Text);
    QTextStream stream(&fh);

    message = new QString(stream.readAll());
    fh.close();

    m_plot_size = screen_size - 80;
    m_app_height = QApplication::screens()[0]->size().height();

    connect(m_a, &QGuiApplication::applicationStateChanged, this, &MainWindow::killTheApp);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAbout);
    connect(closeAction, &QAction::triggered, this, &MainWindow::close);

    coeffs_count_m = 3;
    coeffs_list_m = new CC[3];

    coeffs_list_m[0] = CC(1.0, 0.0);
    coeffs_list_m[1] = CC(0.0, 0.0);
    coeffs_list_m[2] = CC(-1.0, 0.0);

    sols_list = new CC[2];

    m_polish = true;
    m_reverse = false;

    ui->polynomialLabel->setText(formatPoly());

    solver = new PolySolver(coeffs_count_m - 1, coeffs_list_m);
    solver->poly_solve(m_polish, m_reverse, &sols_count, sols_list);
    ui->solutionsLabel->setText(formatSolutions());

    m_im = QApplication::inputMethod();
}

MainWindow::~MainWindow()
{
    delete[] coeffs_list_m;
    delete solver;
    delete[] sols_list;

    delete ui;
}

//  New feature as of September 6, 2020:
//  Perform internal math with 'long double', but
//  return answers as 'double'.
QString MainWindow::formatComplex(CC x)
{
    QString output;
    int prec = 12;
    if (sizeof(CC) == 32) prec = 14;
    if (PolySolver::abs_d(x.imag()) < 1.0e-12) {
        output += QString::number(double(x.real()), 'g', prec);
    } else {
        output += QString("(") + QString::number(double(x.real()), 'g', prec) + " + (" + QString::number(double(x.imag()), 'g', prec) + ")I)";
    }
    return output;
}


QString MainWindow::formatPoly()
{
    QString output;

    for (int i = 1; i < coeffs_count_m - 1; i++) {
        output += formatComplex(coeffs_list_m[i-1]) + " x<sup>" + QString::number(coeffs_count_m - i) + "</sup> &nbsp; + &nbsp; ";
    }
    output += formatComplex(coeffs_list_m[coeffs_count_m - 2]) + " x &nbsp; + &nbsp; ";
    output += QString("(") + formatComplex(coeffs_list_m[coeffs_count_m - 1]) + ") &nbsp; = &nbsp; 0";
    return output;
}

QString MainWindow::formatSolutions()
{
    QString output;
    for (int i = 0; i < sols_count; i++) {
        output += formatComplex(sols_list[i]) + "<br>";
    }
    return output;
}

void MainWindow::UpdateView()
{
    delete solver;
    delete[] sols_list;
    sols_list = new CC[coeffs_count_m - 1];
    solver = new PolySolver(coeffs_count_m - 1, coeffs_list_m);
    solver->poly_solve(m_polish, m_reverse, &sols_count, sols_list);
    ui->polynomialLabel->setText(formatPoly());
    ui->solutionsLabel->setText(formatSolutions());
}

void MainWindow::on_polishBox_stateChanged(int arg1)
{
    if (arg1 == Qt::Checked) {
        m_polish = true;
    } else {
        m_polish = false;
    }
    UpdateView();
}

void MainWindow::on_reverseBox_stateChanged(int arg1)
{
    if (arg1 == Qt::Checked) {
        m_reverse = true;
    } else {
        m_reverse = false;
    }
    UpdateView();
}

void MainWindow::on_newPolyButton_clicked()
{
    DialogWindow *dialog = new DialogWindow(this);
    dialog->showFullScreen();

    ui->newPolyButton->setDisabled(true);

    connect(m_im, &QInputMethod::keyboardRectangleChanged, dialog, &DialogWindow::changeSize);
    connect(dialog, &DialogWindow::reenable_mainw, this, &MainWindow::reenable);
}

void MainWindow::unsetReverse()
{
    if (m_reverse == true)
        ui->reverseBox->setChecked(false);
}

void MainWindow::showAbout()
{
//    QMessageBox *box = new QMessageBox(this);
//    box->setTextFormat(Qt::RichText);
//    box->setText(*message);
//    box->showFullScreen();
    QMessageBox::information(this, QString("Info"), *message);
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
    int t_width = size().width();
    int t_height = size().height();
    ui->verticalLayoutWidget->setGeometry(QRect(5, 45, t_width - 14, t_height - 125));
    QMainWindow::resizeEvent(event);
}

void MainWindow::keyReleaseEvent(QKeyEvent *event) {
    if (event->key() == Qt::Key_Back) {
        event->accept();
        close();
    } else {
        QMainWindow::keyReleaseEvent(event);
    }
}

void MainWindow::reenable() {
    ui->newPolyButton->setDisabled(false);
}

void MainWindow::killTheApp(Qt::ApplicationState astate) {
    if (astate == Qt::ApplicationHidden) {
        (*m_ran) = true;
        showNormal();
    } else if (astate == Qt::ApplicationSuspended && (! *m_ran) && (! m_plotShowing)) {
//        showNormal();           // Courtesy
        close();
    } else if ((astate & Qt::ApplicationSuspended) && m_plotShowing) {
        m_a->exit(RESTART_CODE);
    } else if (astate == Qt::ApplicationSuspended) {
        m_a->exit(0);
    }
    if (astate == Qt::ApplicationActive) {
        (*m_ran) = false;
        showFullScreen();
    }
}
