/* Try to solve a polynomial, using home-grown techniques.
 * By: Dirk Mittler
 * Date: October 25, 2019
 *
 * First Windows-patch added: May 16, 2020
 * Further Windows-patches added until: June 2, 2020
 *
 * Refinement:
 * To deal with the instability of deflation, employ a second pass.
 *
 * The instability of deflation is such, that small, allowed
 * inaccuracies in early roots, leads to successively less accurate,
 * deflated polynomials to approximate. The fact that each of their
 * roots can be found with equally accurate or inaccurate
 * precision, does not make up for the fact that, at that point, the
 * deflated polynomial itself has become inaccurate. Thus, accurate
 * roots were being approximated, for inaccurate polynomials, at first
 * with no way to compensate.
 *
 * Proposed solution:
 * Resume computations with the original polynomial, but this time,
 * Sort the found roots into DESCENDING ORDER, and then,
 * factor out the found roots in reverse order, POLISHING EACH TIME.
 *
 * Refinement first conceived on June 5, 2020 by Dirk Mittler.
 *
 * Major rewrite August 29, 2020:
 * Convert to Object-Oriented form.
 *
 */

#include "rroots_oo_1.h"


void PolySolver::normalize(Polynomial *a)
{
    int array_end = a->offset + a->deg;
    CC first_coeff = a->coeffs[a->offset];
    a->coeffs[a->offset] = 1.0;
    for (int i = a->offset + 1; i <= array_end; i++) {
        a->coeffs[i] /= first_coeff;
    }

    return;
}

//  Apply polynomial to parameter.
CC PolySolver::apply_poly(Polynomial *a, CC param)
{

    int array_end = a->offset + a->deg;

    CC power = 1.0;
    CC result = (a->coeffs)[array_end];
    for (int i = 1; i <= a->deg; i++) {
        power = 1.0;
        for (int j = 1; j <= i; j++) {
            power *= param;
        }

        result += power * (a->coeffs)[array_end - i];
    }

    return result;
}

//  Return real value for real argument, using
//  object's polynomial.
//  New feature as of September 6, 2020:
//  Perform internal math with 'long double', but
//  return answers as 'double'.
double PolySolver::at(double x)
{
    return double(apply_poly(&A, CC(x, 0.0)).real());
}


//  In-place division of polynomial.
//  Returns degree of remaining plynomial in same array.
int PolySolver::divide(Polynomial *a, CC root, CC *remaind)
{
    CC multiple = 0;
    int array_end = a->offset + a->deg;
    for (int i = a->offset; i < array_end; i++) {
        multiple = a->coeffs[i];
        a->coeffs[i+1] += multiple * root;
    }

    *remaind = a->coeffs[array_end];
    a->deg--;

    return a->deg;
}

//  Augmented division of polynomial,
//  may or may not polish.
bool PolySolver::try_root(Polynomial *a, CC *root, bool polish,
                          long double epsilon)
{
    CC adjustment = 0.0;
    CC remaind = 0.0;
    CC deriv = 0.0;

    if (a->deg < 1) return false;
    int N = a->deg;
    double deriv2_a0 = 0.0;

    int passes = 1;
    if (N == 1) polish = false;
    if (polish) passes = 5;
    int array_end = a->offset + N;

    Polynomial quot;
    quot.coeffs = new CC[array_end + 1];

    while (passes-- > 0) {
        *root += adjustment;

        quot.offset = a->offset;
        quot.deg = N;

        for (int i = 0; i <= array_end; i++) {
            quot.coeffs[i] = a->coeffs[i];
        }

        //  (1.0)^0 is possible here...
        deriv2_a0 = N * N * pow(norm(*root) + 1.0, N - 1);

        if (norm(apply_poly(&quot, *root)) > \
                epsilon * epsilon * deriv2_a0) {
            delete[] quot.coeffs;
            return false;
        }

        divide(&quot, *root, &remaind);

        adjustment = CC(0.0, 0.0);

        deriv = -apply_poly(&quot, *root);

        if (polish && norm(deriv) > abs(remaind)) {
            adjustment = remaind / deriv;
        }

    }

    for (int i = 0; i < array_end; i++) {
        a->coeffs[i] = quot.coeffs[i];
    }
    a->deg = quot.deg;
    a->offset = quot.offset;

    delete[] quot.coeffs;

    return true;
}

//  Solver for linear case.
//  Assumes that the equation has been normalized.
inline bool PolySolver::solve_lin(Polynomial *a, CC *root)
{
    if (a->deg != 1) return false;

    *root = - a->coeffs[ a->offset + 1 ] ;
    return true;
}

//  Do the actual work.
bool PolySolver::search_root(Polynomial *a, CC *root, long double epsilon)
{

    int N = a->deg;

    if (N < 1) return false;

    //  !!
    if (N == 1) return solve_lin(a, root);

    CC xa = -1.0;
//	CC xb = exp(CC(0.0, M_PI / (N + 1)));
    CC xb = exp((M_PI / (N + 1)) * 1i);

    CC ya = apply_poly(a, xa);
    CC yb = apply_poly(a, xb);

    double spread = abs(yb - ya);
    double spread2 = 0.0;

    if (N <= 4 && pow(spread, N) < epsilon) {
        return false;
    } else if (N > 4 && pow(spread, N - 2) < N * N * epsilon) {
        return false;
    }

    int misses = 0;

    long double epsilond = epsilon;
    CC t = 0.0;
    CC xt = 0.0;

    //  Really bad form: while (true)
    while (true) {

        epsilond = N * (pow(abs(xa), N - 1)) * epsilon;
        if (epsilond < epsilon / N) epsilond = epsilon / N;

        if (norm(ya) <= epsilond * epsilond) {
            *root = xa;
            return true;
        }

        if (norm(ya - yb) <= epsilon * epsilon * epsilon) {
            return false;
        }

        t = (-ya) / (yb - ya);

        xt = (t * xb) + ((CC(1.0, 0.0) - t) * xa);

        xb = xt + (CC(0.0, 0.5) * (xa - xt));
        xa = xt;

        ya = apply_poly(a, xa);
        yb = apply_poly(a, xb);

        spread2 = abs(ya);

        if (spread2 >= spread) {
            misses++;
        }

        if (misses > N + 2) {
            return false;
        }

        spread = spread2;

    }
}

//  Return number of derivatives...
int PolySolver::create_deriv_list(Polynomial a, Polynomial **l)
{

    CC coeff = 0.0;
    *l = new Polynomial[std::max(a.deg, 1)];

    Polynomial *L = *l;

    int g_offset = a.offset;
    int array_end = a.deg + g_offset;
    L[0].deg = a.deg;
    l[0]->o_deg = a.deg;
    L[0].offset = g_offset;
    L[0].coeffs = new CC[array_end + 1];
    L[0].solved = 0;
    L[0].roots = new CC[a.deg];

    for (int i = 0; i <= array_end; i++) {
        L[0].coeffs[i] = a.coeffs[i];
    }

    if (abs(L[0].coeffs[g_offset]) < MIN_A0) {
        L[0].coeffs[g_offset] = CC(1.0, 0.0);
    }

    normalize(L);

    for (int i = 1; i < a.deg; i++) {
        L[i].deg = L[i-1].deg - 1;
        L[i].o_deg = L[i].deg;
        L[i].offset = g_offset;
        L[i].coeffs = new CC[array_end + 1];
        L[i].solved = 0;
        L[i].roots = new CC[L[i].deg];
        for (int j = 0; j < L[i-1].deg; j++) {

            coeff = L[i-1].coeffs[L[i].deg + g_offset - j];
            L[i].coeffs[L[i].deg + g_offset - j] = coeff * CC(1.0 + j, 0.0);

        }

        normalize(L + i);

    }

    return std::max(a.deg - 1, 0);
}

void PolySolver::destroy_deriv_list(Polynomial *l)
{
    int array_num = l[0].o_deg;
    for (int i = 0; i < array_num; i++) {
        delete[] l[i].coeffs;
        delete[] l[i].roots;
    }

    delete[] l;
    return;
}

bool PolySolver::climb_root(Polynomial *L, CC *root, int i, int mult,
                            long double epsilon)
{
    int s = 0;
    int mult_temp = 0;
    bool last_attempt = false;

    if (i < 0) return true;

    while (( last_attempt = try_root(L + i, root, false, epsilon) )) {
        s = L[i].solved;
        L[i].roots[s] = *root;
        L[i].solved++;
        if (++mult_temp >= mult) break;
    }

    if ( ! last_attempt ) return false;

    return climb_root(L, root, i - 1, mult + 1, epsilon);
}

//  Bubble-Sort roots of solved poly into descending order
void PolySolver::sort_roots(Polynomial *solved_p)
{
    CC temp(0.0, 0.0);
    int deg = solved_p->solved;
    for (int i = 0; i < deg; i++) {
        for (int j = 0; j < i; j++) {
            if (norm(solved_p->roots[j]) < norm(solved_p->roots[i])) {
                temp = solved_p->roots[j];
                solved_p->roots[j] = solved_p->roots[i];
                solved_p->roots[i] = temp;
            }
        }
    }
}

//  Factor out solved roots in reverse order
void PolySolver::reverse_factor(Polynomial *target_p, Polynomial *solved_p,
                                long double epsilon)
{
    int deg = solved_p->solved;
    int offset = target_p->offset;
    if (abs(target_p->coeffs[offset]) < MIN_A0)
        target_p->coeffs[offset] = CC(1.0, 0.0);

    normalize(target_p);

    sort_roots(solved_p);

    CC temp_root = solved_p->roots[--deg];

    while (deg > 0) {
        if (try_root(target_p, &temp_root, true, epsilon)) {
            solved_p->roots[deg--] = temp_root;
        } else {
            solved_p->roots[deg--] = CC(0.0, 0.0);
        }
        temp_root = solved_p->roots[deg];
    }
}


//  Finally, the driver:
//  New feature as of September 6, 2020:
//  Perform internal math with 'long double', but
//  return answers as 'double'.
bool PolySolver::poly_solve(bool polish, bool reverse,
                            int *sol_count, CC *sol_list)
{
    create_deriv_list(A, &PL);
    int s = 0;
    long double epsilon = 1.0E-12;
    Polynomial *L = PL;

    CC current_root = 0.0;
    CC current_conjugate = 0.0;

    if (reverse) {
        polish = true;
    }

    int limits_10 = std::numeric_limits<long double>::digits10 ;

    if (sizeof(CC) == 32) {
        epsilon = 1.0L;
        for (int i = 3 ; i < limits_10 ; i++)
            epsilon /= 10.0L;
    }

    for (int i = A.deg - 1; i >= 0; i--) {
        while ( search_root(L + i, &current_root, epsilon) && \
                try_root(L + i, &current_root, polish, epsilon) ) {

            s = L[i].solved;
            L[i].roots[s] = current_root;
            L[i].solved++;

            climb_root(L, &current_root, i - 1, 2, epsilon);

            if (abs_d(imag(current_root)) > epsilon) {

                current_conjugate = conj(current_root);

                if ( try_root(L + i, &current_conjugate, false, epsilon) ) {

                    s = L[i].solved;
                    L[i].roots[s] = current_conjugate;
                    L[i].solved++;

                    climb_root(L, &current_conjugate, i - 1, 2, epsilon);
                }
            }
        }
    }

    //  New feature as of June 5, 2020:
    //  Factor out solutions in reverse, while
    //  polishing:

    if (reverse)
        reverse_factor(&A, L, epsilon);

    *sol_count = L[0].solved;

    for (int i = 0; i < *sol_count; i++) {
        current_root = L[0].roots[i];

        //  Secondary Polishing...
        if (abs_d(imag(current_root)) <= epsilon) {
            current_root = CC(real(current_root), 0.0);
        }
        if (abs_d(real(current_root)) <= epsilon) {
            current_root = CC(0.0, imag(current_root));
        }

        sol_list[i] = current_root;
    }

    destroy_deriv_list(PL);
    return true;
}

//  Conversion to OO form.
//  Create Solver from coefficients list.
PolySolver::PolySolver(int degree, CC *coeff_list)
{
    PL = NULL;

    A.offset = 0;
    A.deg = degree;

    A.coeffs = new CC[degree + 1];

    for (int c = 0; c <= degree; c++) {
        A.coeffs[c] = coeff_list[c];
    }
}

PolySolver::~PolySolver()
{
    delete[] A.coeffs;
}
